#!/usr/bin/perl

# timtosh -- Tim Newsome <nuisance@cmu.edu>
# You may use any parts of timtosh under the artistic license

use strict;
use vars qw(
    @ISA @EXPORT @EXPORT_OK
    %shellvar %shelldata %command
    );

require Exporter;
use Cwd;
use Term::Cap;
use Sys::Hostname;
use Timtosh::Parser;
use Timtosh::Prompt;
use POSIX qw(
    sys_wait_h signal_h unistd_h termios_h
    );
use Fcntl;

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw();

my @jobs;
my ($fgjob);
my $term;
my $version = "timtosh working version wuzzle";

my %nofork = (
    'cd' => 1,
    'fg' => 1,
    'bg' => 1,
    'exit' => 1,
    'eval' => 1
    );

sub shell {
  while (1) {
    my $command = prompt($term, $shellvar{'PS1'});
    exit if !defined($command);
    my @tokens = tokenize($command) if $command;
    my @expanded = expand(@tokens) if ($#tokens >= 0);
    my @executable = parse(@expanded) if ($#expanded >= 0);
    if ($#executable >= 0) {
      if (execute("check", @executable)) {
	print "Check failed...\n";
      } else {
	execute(undef, @executable);
      }
    }
  }
}

sub chld_handler {
  my $pid;
  POSIX::tcsetpgrp(fileno SHTTY, $$) || die $!;
  while (($pid = waitpid(-1, WNOHANG|WUNTRACED)) > 0) {
    #print STDERR "[$$] SIGCHLD: $pid $?\n";
    if (WIFEXITED($?)) {
      my $jobid = remove_process($pid);
      if (defined($jobid)&& defined($jobs[$jobid])) {
	print STDERR "\n[$jobid] exited with status code ". WEXITSTATUS($?).
	    "\n";
      }
    } elsif (WIFSIGNALED($?)) {
      my $jobid = remove_process($pid);
      if (defined($jobid)&& defined($jobs[$jobid])) {
	print STDERR "\n[$jobid] got killed by signal ". WTERMSIG($?). "\n";
      }
    } elsif (WIFSTOPPED($?)) {
      my ($jobid) = get_job_id($pid);
      if (defined($jobid)&& defined($jobs[$jobid])) {
	print STDERR "\n[$$] $pid got suspended by signal ". WSTOPSIG($?). "\n";
      }
    } else {
      print STDERR "\n[$$] $pid $?\n";
    }
  }
}

#sub tstp_handler {
#  print "[$fgjob] suspending\n";
#  kill SIGSTOP, $jobs[$fgjob]->[0];
#  #open STDIN, $jobs[$fgjob]->[1];
#}

sub init {
  open SHTTY, ">&STDERR" or die $!;

  # Install some signal handlers
#  foreach (keys %SIG) {
#    next if /^(__|INT|TERM|SEGV|KILL|STOP)/;
#    print "Installing handler for $_...\n";
#    $SIG{$_} = sub {
#        POSIX::tcsetpgrp(fileno SHTTY, $$) || die $!;
#	print STDERR " [$$] ", join(" ", @_), "\n";
#      }
#  }

  $SIG{CHLD} = \&chld_handler;
  #$SIG{TSTP} = \&tstp_handler;
  $SIG{TTOU} =
    sub {
      POSIX::tcsetpgrp(fileno SHTTY, $$) || die $!;
      #print STDERR "Grabbing the terminal on TTOU.\n";
    };

  $command{'cd'} = \&cd;
  $command{'exit'} = sub { exit $_[2]; };
  $command{'eval'} = \&evaluate;
  $command{'version'} = \&version;
  $command{'jobs'} = \&jobs;
  $command{'fg'} = \&fg;
  $command{'bg'} = \&bg;

  my @pwent = getpwuid($<);
  $shelldata{'username'} = $pwent[0];
  #$shelldata{'hostname'} = hostname;
  $shelldata{'machinename'} = (split(/\./, $shelldata{'hostname'}))[0];
  $shelldata{'homedir'} = $pwent[7];

  #$shellvar{'PS1'} = '%n@%m:%~%# ';
  $shellvar{'PS1'} = 'timtosh%# ';

  # for curses
  #initscr;
  my $termios = new POSIX::Termios;
  $termios->getattr;
  my $ospeed = $termios->getospeed;
  $term = Tgetent Term::Cap { TERM => undef, OSPEED => $ospeed };
}

sub execute {
  my $check = shift;
  my @data = @{$_[0]};
  my $function = shift @data;
  my @params = @data;

  #print STDERR "[$$] execute $function(", join(", ", @params), ")\n";

  return &$function($check, @params);
}

sub command {
  my $check = shift;
  my $options = shift;
  my @argv = @_;
  my $cmd = $argv[0];

  if (!defined($command{$cmd})) {
    if (($cmd =~ /^\w+$/) && eval "require ShellCommand::$cmd;") {
      # See if we can autoload this fu.
      $command{$cmd} = \&exec_module;
      return 0 if $check;

    } else {
      $command{$cmd} = \&exec_command;
      # TODO: Actually do a check here.
      return 0 if $check;
    }

  } else {
    return 0 if $check;
  }

  my $function = $command{$cmd};

  my $option;
  foreach $option (@{$options}) {
    #print "option $option\n";
    if ($option->[0] eq 'REDIRECT') {
      my $handle = $option->[1];
      my $type = $option->[2];
      my $file = $option->[3];
      if ($handle == 0) {
	open STDIN, "$type $file" || die;
      } elsif ($handle == 1) {
	open STDOUT, "$type $file" || die;
      } elsif ($handle == 2) {
	open STDERR, "$type $file" || die;
      }
    } else {
      die "Unknown option: $option->[0]";
    }
  }

  return &$function(@argv);
}

sub exec_complex_command {
  my $check = shift;
  my @commands = @{$_[0]};
  shift;
  my $foreground = shift;

  my @processes;

  my $pgrp_leader;

  # reset terminal back to normal
  #noraw;
  #nl;
  #nonl;
  #echo;
  #endwin; 

  if ($check) {
    # check if all commands are valid, and load up any .pm fu
    my $cmd;
    foreach $cmd (@commands) {
      return 1 if execute("check", $cmd);
    }
    return 0;
  }

  {
    # Fire up the first command in the stream.
    my $cmd = shift @commands;

    # this check is kind of a hack
    if ($cmd->[0] eq \&command && $nofork{$cmd->[2]}) {
      # no need to fork
      return execute(undef, $cmd);
    }

    # make a pipe to which the child can write and the parent can read from
    if ($#commands >= 0) {
      pipe READ, WRITE;
    }

    if ($pgrp_leader = fork()) {
      # parent
      if ($#commands >= 0) {
	close WRITE;
	open INPUT, "<&READ";
      }

    } else {
      # child
      close READ;

      # make a new process group
      POSIX::setpgid(0, $$) or die $!;

      if ($foreground) {
	# Make sure ISIG is set
	my $termios = POSIX::Termios->new;
	$termios->getattr(fileno SHTTY);
	my $lflags = $termios->getlflag();
	$lflags |= ISIG;
	$termios->setlflag($lflags);
	$termios->setattr(fileno SHTTY, TCSADRAIN);

	# give this process the tty
	POSIX::tcsetpgrp(fileno SHTTY, $$) or die $!;
      }

      if ($#commands >= 0) {
	open STDOUT, ">&WRITE";
      }

      exit execute(undef, $cmd);
    }

    push @processes, [ $pgrp_leader ];
  }

  while ($#commands >= 0) {
    my $cmd = shift @commands;

    my $child_pid;

    if ($#commands >= 0) {
      pipe READ, WRITE;
    }

    if ($child_pid = fork) {
      # parent
      if ($#commands >= 0) {
	close WRITE;
	open INPUT, "<&READ";
      }

      push @processes, [ $child_pid ];

    } else {
      # child
      close READ;

      # join the process group of the first process
      POSIX::setpgid(0, $pgrp_leader) or die $!;

      open STDIN, "<&INPUT";
      if ($#commands >= 0) {
	open STDOUT, ">&WRITE";
      }

      exit execute(undef, $cmd);
    }
  }

  my $jid = add_job(@processes);

  if ($foreground) {
    return waitforjob($jid);
  }

  return 0;
}

sub exec_command {
  #print "exec_command(", join(", ", @_), ")\n";
  my $bin = $_[0];

  exec $bin @_ or die $!;
}

sub exec_module {
  my $function;
  eval "\$function = \\&ShellCommand::$_[0]::execute;";

  exit &$function(@_);
}

sub sequentially {
  my $check = shift;
  my $cmd;
  foreach $cmd (@_) {
    execute $check, $cmd;
  }
}

sub cd {
  my $target = $_[1];

  if ($target eq '') {
    $target = $ENV{'HOME'};
  }
  if (chdir $target) {
    $ENV{'PWD'} = cwd();
  } else {
    print "cd: $!\n";
  }
}

sub evaluate {
  shift;
  eval join(" ", @_);
  print $@ if $@;
}

sub version {
  print "$version\n";
}

sub get_job_id {
  my $pid = shift;

  my $i;

  for ($i = 0; $i <= $#jobs; $i++) {
    next unless defined $jobs[$i];
    my @processes = @{$jobs[$i]};

    my $j;
    for ($j = 0; $j <= $#processes; $j++) {
      if ($processes[$j]->[0] eq $pid) {
	return ($i, $j);
      }
    }
  }

  return undef;
}

sub jobs {
  for (my $j = 0; $j <= $#jobs; $j++) {
    next unless defined $jobs[$j];
    my @data = @{$jobs[$j]};
    my @pid;
    foreach (@data) {
      push @pid, $_->[0];
    }
    print "  [$j] pids: ", join(", ", @pid), "\n";
  }
}

sub add_job {
  my @data = @_;
  my $i;

  for ($i = 0; $i <= $#jobs; $i++) {
    last if (!defined($jobs[$i]));
  }

  $jobs[$i] = [ @data ];

  return $i;
}

sub remove_process {
  my $pid = shift;
  my ($jid, $pindex) = get_job_id($pid);

  return undef unless (defined $jid && defined $pindex);

  my @processes = @{$jobs[$jid]};
  splice @processes, $pindex, 1;

  if ($#processes < 0) {
    undef $jobs[$jid];
  }

  return $jid;
}

sub remove_job {
  my $jid = shift;
  undef $jobs[$jid];
  return $jid;
}

sub fg {
  shift;
  my $jobid = shift;
  $jobid =~ s/[^\d]+//;

  my @pids = @{$jobs[$jobid]};

  POSIX::tcsetpgrp(fileno SHTTY, $pids[0]->[0]) or die $!;

  my $pid;
  foreach $pid (reverse @pids) {
    kill SIGCONT, $pid->[0]; 
  }

  waitforjob($jobid);
}

sub bg {
  shift;
  my $jobid = shift;
  $jobid =~ s/[^\d]+//;

  my @pids = @{$jobs[$jobid]};

  my $pid;
  foreach $pid (reverse @pids) {
    kill SIGCONT, $pid; 
  }
}

sub waitforjob {
  my $jid = shift;

  my @data = @{$jobs[$jid]};
  waitpid $data[$#data]->[0], WUNTRACED;

  POSIX::tcsetpgrp(fileno SHTTY, $$) || die $!;

  if (WIFEXITED($?)) {
    remove_job($jid);
    #print "$pid exited - ". WEXITSTATUS($?). "\n";
    return WEXITSTATUS($?);
  } elsif (WIFSTOPPED($?)) {
    print $data[$#data]->[0], " stopped by ". WSTOPSIG($?). "\n";
    return undef;
  } elsif (WIFSIGNALED($?)) {
    remove_process($data[$#data]->[0]);
    #print "$pid killed - ". WTERMSIG($?). "\n";
    return -WTERMSIG($?);
  } else {
    die "WTF!";
  }
}

init();
shell();

1;
